// -*- mode:c++;tab-width:2;indent-tabs-mode:t;show-trailing-whitespace:t;rm-trailing-spaces:t -*-
// vi: set ts=2 noet:
//
// (c) Copyright Rosetta Commons Member Institutions
// (c) This file is part of the Rosetta software suite and is made available under license.
// (c) The Rosetta software is developed by the contributing members of the Rosetta Commons.
// (c) For more information, see http://www.rosettacommons.org. Questions about this can be
// (c) addressed to University of Washington UW TechTransfer, email: license@u.washington.edu.

/// @brief EDA-based AbRelax application
///
/// ---------------------------------------
/// @author David Simoncini

// keep these headers first for compilation with Visual Studio C++
#include <utility/io/izstream.hh>
// #include <utility/pointer/boost/owning_ptr.hh>
// #include <utility/pointer/boost/access_ptr.hh>
// #include <utility/pointer/boost/ReferenceCount.hh>

#include <protocols/jobdist/JobDistributors.hh>
#include "EdaJobDistributor.hh"
#include <protocols/jobdist/Jobs.hh>

// Unit Headers
#include "EdaAbrelax.hh"
#include "EdaAbinitio.hh"
#include "EdaPose.hh"
#include <protocols/abinitio/AbrelaxApplication.hh>
#include <protocols/constraints_additional/AdditionalConstraintCreators.hh>
#include <protocols/loops/util.hh>
#include <protocols/loops/Loops.hh>
#include <protocols/loops/Loop.hh>
#include <protocols/loops/loop_mover/refine/LoopMover_KIC.hh>
#include <protocols/loops/loops_main.hh>

#if defined(WIN32) || defined(__CYGWIN__)
	#include <ctime>
#endif

// Package Headers
#include <core/kinematics/util.hh>
#include <core/kinematics/MoveMap.hh>
#include <protocols/abinitio/ClassicAbinitio.hh>
#include <protocols/abinitio/MembraneAbinitio.hh>
#include <protocols/abinitio/FoldConstraints.hh>
#include <protocols/abinitio/KinematicTaskControl.hh>
#include <protocols/abinitio/LoopJumpFoldCst.hh>
#include <protocols/abinitio/DoubleLayerKinematicAbinitio.hh>
#include <protocols/abinitio/Templates.hh>
#include <protocols/abinitio/TemplateJumpSetup.hh>
#include <protocols/abinitio/PairingStatistics.hh>
#include <protocols/abinitio/StrandConstraints.hh>
#include <protocols/simple_moves/FragmentMover.hh>
#include <protocols/abinitio/Protocol.hh>
#include <protocols/idealize/IdealizeMover.hh>
#include <protocols/jumping/SheetBuilder.hh>
#include <protocols/jumping/RandomSheetBuilder.hh>
#include <protocols/jumping/JumpSetup.hh>
#include <core/scoring/dssp/PairingsList.hh>
#include <protocols/jumping/ResiduePairJumpSetup.hh>
#include <core/fragment/SecondaryStructure.hh>
#include <core/scoring/dssp/StrandPairing.hh>
#include <protocols/jumping/util.hh>
#include <protocols/jumping/MembraneJump.hh>
#include <protocols/evaluation/PoseEvaluator.hh>
#include <protocols/evaluation/EvaluatorFactory.hh>
#include <protocols/evaluation/PCA.hh>
#include <core/pose/datacache/CacheableDataType.hh>
#include <basic/datacache/BasicDataCache.hh>
#include <core/scoring/SS_Killhairpins_Info.hh>
// AUTO-REMOVED #include <core/scoring/SS_Killhairpins_Info.fwd.hh>
#include <core/scoring/methods/ContactOrderEnergy.hh>

// Project Headers
#include <core/types.hh>
#include <core/conformation/Conformation.hh>
#include <core/pose/Pose.hh>
#include <core/pose/metrics/CalculatorFactory.hh>
#include <core/pose/util.hh>
#include <basic/MetricValue.hh>
#include <basic/options/option.hh>
#include <basic/options/keys/eda.OptionKeys.gen.hh>
#include <basic/options/keys/abinitio.OptionKeys.gen.hh>
#include <basic/options/keys/constraints.OptionKeys.gen.hh>
// AUTO-REMOVED #include <basic/options/keys/evaluation.OptionKeys.gen.hh>
#include <basic/options/keys/filters.OptionKeys.gen.hh>
#include <basic/options/keys/frags.OptionKeys.gen.hh>
#include <basic/options/keys/in.OptionKeys.gen.hh>
#include <basic/options/keys/jumps.OptionKeys.gen.hh>
#include <basic/options/keys/loopfcst.OptionKeys.gen.hh>
#include <basic/options/keys/loops.OptionKeys.gen.hh>
#include <basic/options/keys/out.OptionKeys.gen.hh>
#include <basic/options/keys/run.OptionKeys.gen.hh>
#include <basic/options/keys/templates.OptionKeys.gen.hh>
#include <basic/options/keys/relax.OptionKeys.gen.hh>
#include <basic/options/keys/score.OptionKeys.gen.hh>
#include <core/fragment/BBTorsionSRFD.hh>
#include <core/fragment/FrameIteratorWorker_.hh>
#include <core/fragment/util.hh>
#include <core/fragment/FragmentIO.hh>
#include <core/chemical/ChemicalManager.hh>
#include <core/io/silent/SilentFileData.hh>
#include <core/scoring/Energies.hh>
#include <core/scoring/constraints/ConstraintFactory.hh>
#include <core/scoring/constraints/ConstraintSet.hh>
#include <core/scoring/constraints/ConstraintIO.hh>
#include <core/scoring/func/HarmonicFunc.hh>
#include <core/scoring/constraints/util.hh>
#include <protocols/toolbox/pose_metric_calculators/ClashCountCalculator.hh>
// AUTO-REMOVED #include <core/io/pdb/pose_io.hh>
#include <core/io/silent/silent.fwd.hh>
#include <core/io/silent/SilentStructFactory.hh>
#include <core/io/raw_data/DisulfideFile.hh>
#include <core/scoring/ScoreFunction.fwd.hh>
#include <core/scoring/ScoreType.hh>
#include <core/scoring/ScoreFunctionFactory.hh>
#include <core/scoring/func/Func.hh>
#include <core/scoring/constraints/BoundConstraint.hh>
#include <core/scoring/MembraneTopology.hh>
#include <core/sequence/util.hh>
#include <core/sequence/Sequence.hh>
#include <protocols/simple_filters/RmsdEvaluator.hh>
#include <protocols/simple_filters/JumpEvaluator.hh>
#include <protocols/evaluation/TimeEvaluator.hh>
#include <protocols/constraints_additional/ConstraintEvaluator.hh>
#include <protocols/simple_filters/PoseMetricEvaluator.hh>
#include <protocols/evaluation/util.hh>
// AUTO-REMOVED #include <protocols/evaluation/ChemicalShiftEvaluator.hh>
#include <protocols/loops/loop_closure/ccd/SlidingWindowLoopClosure.hh>
#include <protocols/loops/loop_closure/ccd/WidthFirstSlidingWindowLoopClosure.hh>
#include <protocols/loops/loop_closure/ccd/FASelectSlidingWindowLoopClosure.hh>
#include <protocols/loops/loop_mover/LoopMover.hh>
#include <protocols/loops/Exceptions.hh>
#include <protocols/filters/Filter.hh>
#include <protocols/simple_filters/RGFilter.hh>
#include <protocols/simple_filters/COFilter.hh>
#include <protocols/simple_filters/SheetFilter.hh>
#include <protocols/simple_filters/PDDFScoreFilter.hh>
#include <protocols/simple_filters/SAXSScoreFilter.hh>
#include <protocols/moves/MoverStatus.hh>
#include <protocols/simple_moves/RepulsiveOnlyMover.hh>

//numeric headers
#include <numeric/random/random.hh>

// Utility headers
#include <utility/string_util.hh>
#include <utility/vector1.hh>
#include <basic/Tracer.hh>
#include <utility/io/util.hh>
#include <utility/exit.hh>

// C++ headers
#include <cstdlib>
// AUTO-REMOVED #include <iterator>
#include <string>
#include <utility>
#include <vector>

#include <core/fragment/FragData.hh>
#include <core/fragment/Frame.hh>
#include <core/fragment/FrameIterator.hh>
#include <core/fragment/FrameList.hh>
#include <core/id/SequenceMapping.hh>
#include <core/import_pose/import_pose.hh>
#include <core/pose/annotated_sequence.hh>
#include <core/scoring/func/FuncFactory.hh>
#include <core/util/SwitchResidueTypeSet.hh>
#include <protocols/abinitio/KinematicAbinitio.hh>
#include <protocols/relax/ClassicRelax.hh>
#include <protocols/relax/FastRelax.hh>
#include <protocols/relax/util.hh>
#include <utility/vector0.hh>
#include <ObjexxFCL/format.hh>

//Auto using namespaces
namespace ObjexxFCL { } using namespace ObjexxFCL; // AUTO USING NS
namespace ObjexxFCL { namespace fmt { } } using namespace ObjexxFCL::fmt; // AUTO USING NS

static basic::Tracer tr("EdaAbrelaxApplication");
//static numeric::random::RandomGenerator RG(423465);  // <- Magic number, do not change it!

namespace protocols {
namespace abinitio {

using core::Size;
using namespace std;
using namespace core;
using namespace protocols;
using namespace fragment;
using namespace abinitio;
using namespace jumping;
using namespace evaluation;
using namespace basic::options;
using namespace basic::options::OptionKeys;

	EdaAbrelax::EdaAbrelax():AbrelaxApplication()
{
	
}

EdaAbrelax::~EdaAbrelax() {}

/// @details Shallow copy to mimic the pre 9/8/09 compiler-generated version of this
/// method.  If you add new
EdaAbrelax::EdaAbrelax( EdaAbrelax const & src ):AbrelaxApplication(src)
{
}


////////////////////////////////////////////////////////////////////////////////////////////////////
///@detail setup_fold() all initialization that is necessary to run abinitio production loop in fold()
/// read fragments, make pose from sequence, get constraints, set jumps, movemap .etc
/// the two parameters are OUTPUT:
///    extended_pose to run A) with ( might actually contain native starting structure (option!) )
///    prot_ptr: an initialized instance of ClassicAbinitio, FoldConstraints or JumpingFoldConstraints depending on
///   user settings.

void EdaAbrelax::generate_extended_pose( EdaPose &extended_pose, std::string const& sequence ) const {
	Parent::generate_extended_pose(extended_pose, sequence);
	extended_pose.resize_frag_keys();
}

void EdaAbrelax::setup_fold( EdaPose& extended_pose, EdaAbinitioOP& prot_ptr ) {
	using namespace basic::options;
	using namespace basic::options::OptionKeys;
	// ==========================================================================
	///  --------- fold()-specific setup --------------------------
	// ==========================================================================

	// ---------------------------------------------------------------------------------------------------------------
	// initialize pose
	generate_extended_pose( extended_pose, sequence_ );
	// apply a mover which calculates only repulsive energy on designate residues
	protocols::simple_moves::RepulsiveOnlyMover replonly;
	replonly.apply( extended_pose );

	if ( option[ OptionKeys::abinitio::start_native ]() ) {
		copy_native_structure( extended_pose );
	} else if ( option[ in::file::s ].user() ) {
		EdaPoseOP tmp_pose( new EdaPose );
		std::string fn = option[ in::file::s ](1);
		core::import_pose::pose_from_file( *tmp_pose, fn, core::import_pose::PDB_file );
		copy_structure( extended_pose, *tmp_pose );
	}

	// Fix disulfides if a file is given
	if ( option[ basic::options::OptionKeys::in::fix_disulf ].user() ) {
		io::raw_data::DisulfideFile ds_file( option[ OptionKeys::in::fix_disulf ]() );
		utility::vector1< std::pair<Size,Size> > disulfides;
		ds_file.disulfides(disulfides, extended_pose);
		extended_pose.conformation().fix_disulfides( disulfides );
	}

	// ---------------------------------------------------------------------------------------------------------------
	bRelax_ = option[ OptionKeys::abinitio::relax ]() ||
		option[ OptionKeys::abinitio::fastrelax ]() ||
		option[ OptionKeys::abinitio::multifastrelax ]();

	// FRAGMENTS:
	// stores in FragSetOP fragset_large_, fragset_small_;
	setup_fragments();
	// read in constraint_forest, generate random sample
	initialize_constraint_forest( extended_pose );

	// add constraints if available
	add_constraints( extended_pose );

	// make a MoveMap
	kinematics::MoveMapOP movemap(new	kinematics::MoveMap);
	movemap->set_bb( true );

	if ( option[ OptionKeys::abinitio::fix_residues_to_native ].user() ) {
		utility::vector1< int> const& fix_start_ends( option[ OptionKeys::abinitio::fix_residues_to_native ]() );
		for ( Size i=1; i + 1 <= fix_start_ends.size(); i+=2 ) {
			Size const start( fix_start_ends[ i ]);
			Size const end( fix_start_ends[ i + 1 ]);
			if ( !(end >= start) ) utility_exit_with_message("end < start in abinitio:fix_residues_to_native");
			fragment::Frame long_frame(start, end-start+1 );
			//create apropriate length FragData object
			FragData frag( SingleResidueFragDataOP(new BBTorsionSRFD), end-start+1 );

			// get torsion angles from native pose
			frag.steal( *native_pose_, long_frame );

			// apply native torsions to extended structue
			frag.apply( extended_pose, long_frame );

			for ( Size pos = start; pos <= end; pos++ ) {
				movemap->set_bb( pos, false);
			}
		}
	}

	EdaAbinitioOP eda_ab(new EdaAbinitio( fragset_small_, fragset_large_, movemap ));
	prot_ptr = eda_ab;

	EdaAbinitio& abinitio_protocol( *prot_ptr ); // hide the fact that protocol is a pointer
	
	/// initialize protocol
	abinitio_protocol.init( extended_pose );
	if ( option[ OptionKeys::loopfcst::use_general_protocol ] ) {
		abinitio_protocol.return_centroid( !(bRelax_ || option[ OptionKeys::abinitio::return_full_atom ]) );
	} else {
		abinitio_protocol.return_centroid( true );
	}
	if ( evaluator_->size() ) abinitio_protocol.set_evaluation( evaluator_ );
}

void EdaAbrelax::fold( EdaPose &init_pose, EdaAbinitioOP prot_ptr ) {
	EdaAbinitio& abinitio_protocol( *prot_ptr ); // hide the fact that protocol is a pointer
	
	using namespace basic::options;
	using namespace basic::options::OptionKeys;
	using namespace core::scoring::constraints;
	using protocols::jobdist::BasicJob;
	using protocols::jobdist::BasicJobOP;
	using protocols::jobdist::PlainSilentFileJobDistributor;
	// determine nstruct
  int const nstruct = std::max( 1, option [ out::nstruct ]() / option [ eda::nb_iterations ] );

	// setup JobDistributor stuff
	utility::vector1< BasicJobOP > input_jobs;
	option.add_relevant( OptionKeys::out::pdb );
	if ( option[ in::file::tags ].user() ) {
		// get input tags
		typedef utility::vector1< std::string > TagList;
		TagList input_tags;
		utility::io::izstream tag_file( option[ in::file::tagfile ]() );

		std::copy( std::istream_iterator< std::string >( tag_file ), std::istream_iterator< std::string >(),
			std::back_inserter( input_tags ) );

		// create jobs
		for ( TagList::const_iterator it = input_tags.begin(), eit = input_tags.end(); it!=eit; ++it ) {
			BasicJobOP job (new BasicJob( *it, "resample", nstruct));
			tr.Debug << "create resample job" << *it << std::endl;
			input_jobs.push_back( job );
		}
	} else { //default behaviour
		BasicJobOP job (new BasicJob("" /*no input tag*/, "abinitio_relax", nstruct));
		input_jobs.push_back( job );
	}

	EdaJobDistributor jobdist( input_jobs );
	BasicJobOP curr_job;
	int curr_nstruct;
	jobdist.startup();
	bool bEndrun = false;
	// setup scorefunctions
	// this is called in setup_fold: add_constraints( extended_pose ); //such that scorefxn setup knows about constraints...
	core::scoring::ScoreFunctionOP centroid_scorefxn( generate_scorefxn( false /*fullatom*/ ) );
	core::scoring::ScoreFunctionOP fullatom_scorefxn( generate_scorefxn( true /*fullatom*/ ) ); // this is the score12 standard score function ?! );
	abinitio_protocol.set_fullatom_scorefxn( fullatom_scorefxn );
	abinitio_protocol.set_centroid_scorefxn( centroid_scorefxn );

	evaluation::TimeEvaluatorOP run_time( NULL );
	if ( !option[ OptionKeys::run::no_prof_info_in_silentout ]() ) {
		add_evaluation( run_time = evaluation::TimeEvaluatorOP(new evaluation::TimeEvaluator )); //just don't use this in integration tests!
		abinitio_protocol.set_evaluation( evaluator_ );
	}

	for (unsigned nb_iter=0; nb_iter<option [ eda::nb_iterations ];nb_iter++) {
	// production loop
	while ( jobdist.next_job(curr_job, curr_nstruct) && !bEndrun ) {
		time_t pdb_start_time = time(NULL);
		if ( run_time && !option[ OptionKeys::abinitio::no_write_failures ]() )	{	 //if we omit decoys we want to count from write-event to write-
			run_time->reset(); //reset clock of TimeEvaluator
		}

	#ifdef BOINC
		std::cerr << "Starting work on structure: " << curr_job->output_tag(curr_nstruct) << std::endl;
	#endif

		// retrieve starting pose
		EdaPose fold_pose ( init_pose );
//membrane jumping set up the proper foldtree
		// if( membrane_jumps_ && membrane_jumps_->defined() ) {
		// 	Size njumps = option[jumps::njumps]();
		// 	membrane_jumps_->setup_fold_tree(fold_pose,njumps);
		// }

		// kill hairpins
		//		using core::pose::datacache::CacheableDataType::SS_KILLHAIRPINS_INFO;
		if ( option[ OptionKeys::abinitio::kill_hairpins ].user() ) {
			fold_pose.data().set( core::pose::datacache::CacheableDataType::SS_KILLHAIRPINS_INFO, basic::datacache::DataCache_CacheableData::DataOP(new core::scoring::SS_Killhairpins_Info));
			runtime_assert( fold_pose.data().has( core::pose::datacache::CacheableDataType::SS_KILLHAIRPINS_INFO ) );
			core::scoring::SS_Killhairpins_Info & hairpins=*( utility::pointer::static_pointer_cast< core::scoring::SS_Killhairpins_Info > ( fold_pose.data().get_ptr( core::pose::datacache::CacheableDataType::SS_KILLHAIRPINS_INFO ) ));

			hairpins.setup_killhairpins();
		}

	#ifdef BOINC_GRAPHICS
		// attach boinc graphics pose observer
		protocols::boinc::Boinc::attach_graphics_current_pose_observer( fold_pose );
	#endif

		if ( bRelax_ ) {  //always add all f@@#$ columns so we never-ever have a column mismatch....
			protocols::relax::ClassicRelax().setPoseExtraScore( fold_pose );
		}

		// need to save RG states such that choices for constraints and fold-tree are the same.
		if( ! abrelax_checkpoints_.recover_checkpoint( fold_pose, jobdist.get_current_output_tag(), "rg_state") ){
			abrelax_checkpoints_.checkpoint( fold_pose, jobdist.get_current_output_tag(), "rg_state");
		}
		abrelax_checkpoints_.debug( jobdist.get_current_output_tag(), "rg_state", numeric::random::rg().uniform() );

		// perturb phi/psi randomly -- should be different each run
		if ( option[ OptionKeys::abinitio::perturb ].user() ) {
			Real sig = option[ OptionKeys::abinitio::perturb ];
			for ( Size pos = 1; pos <= fold_pose.total_residue(); pos++ ) {
				fold_pose.set_phi( pos, fold_pose.phi( pos ) + numeric::random::gaussian()*sig );
				fold_pose.set_psi( pos, fold_pose.psi( pos ) + numeric::random::gaussian()*sig );
				fold_pose.set_omega( pos, fold_pose.omega( pos ) );
			}
		}

		// run abinitio

		// set the TAG
		std::ostringstream oss;
		oss << nb_iter;
		abinitio_protocol.set_current_tag( jobdist.get_current_output_tag()+"_"+oss.str() );

		// and the JOB (that's different)
		abinitio_protocol.set_current_job( curr_job );
		tr.Debug << "fold_pose is " << (fold_pose.is_fullatom() ? " fullatom " : " centroid " ) << "before protocol run "<<std::endl;

		std::string output_tag = abinitio_protocol.get_current_tag();
		abinitio_protocol.apply( fold_pose );
		bool loop_closure_failed( false ); //!fold_pose.fold_tree().num_cutpoint() );

		if ( option[ OptionKeys::abinitio::close_loops ]()
			&& abinitio_protocol.get_last_move_status() == moves::MS_SUCCESS
			&& !option[ OptionKeys::loopfcst::use_general_protocol ] )
		{
			tr.Info << "OLD PATHWAY: close loops" << std::endl;
			loop_closure_failed = !close_loops( fold_pose, centroid_scorefxn, jobdist.get_current_output_tag() );
		}

		if ( bRelax_ && option [ OptionKeys::abinitio::debug ] ) {
			io::silent::SilentFileData outsfd;
			std::string silent_file = option[ basic::options::OptionKeys::out::file::silent ]() + "_" + "before_relax";

			io::silent::SilentStructOP pss = io::silent::SilentStructFactory::get_instance()->get_silent_struct_out();
			process_decoy( fold_pose, fold_pose.is_fullatom() ? *fullatom_scorefxn : *centroid_scorefxn, jobdist.get_current_output_tag(), *pss );
			outsfd.write_silent_struct( *pss, silent_file );
		}

		// run relax if applicable. Use filters to decide which structures to relax and which not to relax!
		bool passes_filters = check_filters( fold_pose );
		if(option[OptionKeys::abinitio::membrane]){
			passes_filters=true;
		}
		bool bProcessDecoy( true );
			// run relax if applicable
		// don't relax if we failed filters or loop_closing, or if option[ relax_with_jumps ] is true
		bool bCanRelax = abinitio_protocol.get_last_move_status() == moves::MS_SUCCESS;
		if( option[ basic::options::OptionKeys::abinitio::relax_failures ]() ) bCanRelax = true;
		bCanRelax = bCanRelax && passes_filters
			&& ( !loop_closure_failed || option[ OptionKeys::abinitio::relax_with_jumps ]() );

		if ( bRelax_ ) {
			if ( !fold_pose.is_fullatom() ) {
				EdaPose const centroid_pose ( fold_pose );
				ResolutionSwitcher res_switch( centroid_pose, false, true, true );

				if ( option[ OptionKeys::constraints::cst_fa_file ].user() ) 	 res_switch.set_map_cst_from_centroid_to_fa( false ); //will override any attempt to map centroid constraints to full-atom constraints -- use user-defined file instead!

				res_switch.apply( fold_pose );

				if ( option[ OptionKeys::constraints::cst_fa_file ].user() ) 	{
					ConstraintSetOP cstset_ = ConstraintIO::get_instance()->read_constraints( get_cst_fa_file_option(), ConstraintSetOP( new ConstraintSet ), fold_pose );
					fold_pose.constraint_set( cstset_ );
				}
			}

			if( option[ basic::options::OptionKeys::abinitio::close_loops_by_idealizing ]() ){
				// record cutpoints
				protocols::loops::LoopsOP cloops(new protocols::loops::Loops());
				for ( Size ncut = 1; ncut <= (Size) fold_pose.fold_tree().num_cutpoint(); ncut++ ) {
					Size cutpoint = fold_pose.fold_tree().cutpoint( ncut );
					protocols::loops::Loop newloop (
						 std::max( (int) 1, int(cutpoint - 5) ),
						 std::min( (int) fold_pose.total_residue(), int(cutpoint + 5) ),
						 0
					);

					if( cloops->size() >= 2 )
					if( newloop.start() <= ( *cloops )[cloops->size()-1].stop() ) newloop.set_start( ( *cloops )[cloops->size()-1].stop() +2 );
					newloop.choose_cutpoint( fold_pose );
					cloops->add_loop( newloop );
				}

				cloops->auto_choose_cutpoints( fold_pose );

				// forget about foldtree & cuts
				core::kinematics::FoldTree f_new;
				f_new.simple_tree( fold_pose.total_residue() );
				fold_pose.fold_tree( f_new );

				//idealize
				protocols::idealize::IdealizeMover idealizer;
				idealizer.fast( false );
				idealizer.apply( fold_pose );

				relax( fold_pose, fullatom_scorefxn, jobdist.get_current_output_tag() );

				if( option[ basic::options::OptionKeys::abinitio::optimize_cutpoints_using_kic ]() ){
					protocols::loops::fold_tree_from_loops( fold_pose, *cloops, f_new, true /* include terminal cutpoints */);
					fold_pose.fold_tree( f_new );
					core::scoring::ScoreFunctionOP refine_scorefxn = fullatom_scorefxn->clone();
					protocols::loops::loop_mover::refine::LoopMover_Refine_KIC refine_kic( cloops, refine_scorefxn );
					refine_kic.apply( fold_pose );

					// Return fold tree to norml state
					f_new.simple_tree( fold_pose.total_residue() );
					fold_pose.fold_tree( f_new );
				}
			}

			if ( bCanRelax ) {
				if ( option[ basic::options::OptionKeys::abinitio::multifastrelax ]() ) {
					bEndrun = multi_fast_relax( abinitio_protocol, fullatom_scorefxn, jobdist, curr_nstruct, curr_job );
					bProcessDecoy = false;
					if ( bEndrun ) break;
				} else {
					relax( fold_pose, fullatom_scorefxn, jobdist.get_current_output_tag() );
				}
			} else { //cannot relax
				(*fullatom_scorefxn)( fold_pose );
				if ( option[ basic::options::OptionKeys::abinitio::fastrelax ]() ) {
				} else {
					protocols::relax::ClassicRelax().setPoseExtraScore( fold_pose ); // ClassicRelax adds four columns
				}
			}
		} // if ( bRelax_ )

		//Add contact order to score file as an extra column
		core::scoring::methods::ContactOrderEnergy co_energy;
		Real contact_order = co_energy.calculate_contact_order( fold_pose );
		core::pose::setPoseExtraScore( fold_pose, "co", contact_order );

		// process decoy if this hasn't happened yet
		if ( bProcessDecoy ) {
			if( option[ run::checkpoint ]() ){
				core::pose::setPoseExtraScore( fold_pose, "ichkpnt",
						abinitio_protocol.get_checkpoints().get_checkpoint_recoveries() +
						abrelax_checkpoints_.get_checkpoint_recoveries()     );
			}

			// analyze result
			io::silent::SilentFileData outsfd;

			//make sure that number of columns does not change -- ever
			outsfd.strict_column_mode( true );

			io::silent::SilentStructOP pss = io::silent::SilentStructFactory::get_instance()->get_silent_struct_out();

			// abinitio produces n_stored structures -- the last one is the same as the final structure n_stored.
			//			Size n_stored( abinitio_protocol.structure_store().size() );
			std::string new_output_tag ( output_tag );
			if ( !passes_filters  && loop_closure_failed ) {
				new_output_tag = "X_"+new_output_tag.substr(2);
			} else if( loop_closure_failed ) {
				new_output_tag = "C_"+new_output_tag.substr(2);
			} else if( !passes_filters ) {
				new_output_tag = "F_"+new_output_tag.substr(2);
			} else if ( abinitio_protocol.get_last_move_status() != moves::MS_SUCCESS ) {
				new_output_tag = "P_"+new_output_tag.substr(2);
			}

			if ( !option[ OptionKeys::abinitio::no_write_failures ]()
				|| ( passes_filters && !loop_closure_failed && abinitio_protocol.get_last_move_status() == moves::MS_SUCCESS ) ) {
				// write to silent file
				process_decoy( fold_pose, fold_pose.is_fullatom() ? *fullatom_scorefxn : *centroid_scorefxn, new_output_tag, *pss );
				// write this to score-file if applicable
				if ( silent_score_file_ ) {
					silent_score_file_ -> write_silent_struct( *pss,  silent_score_file_->filename(), true /* bWriteScoresOnly */ );
				}
				if ( option[ OptionKeys::out::pdb ] ) { fold_pose.dump_pdb( std::string(option[ OptionKeys::out::path::path ]())  + "/" + new_output_tag + ".pdb");
					outsfd.add_structure( pss ); 
				}
				if ( run_time && option[ OptionKeys::abinitio::no_write_failures ]() )	{	 //if we omit decoys we want to count from write-event to write-
					run_time->reset(); //reset clock of TimeEvaluator
				}

			}

			jobdist.dump_silent( outsfd ); // does the same thing as: outsfd.write_all( filename ); //cool bulk-writing makes clusters happy
		} //bProcessDecoy ( false if multi-fast_relax )
		// clean up
		jobdist.store_score_and_frags(fold_pose,output_tag,fold_pose.is_fullatom() ? *fullatom_scorefxn : *centroid_scorefxn);
		time_t pdb_end_time = time(NULL);
		tr.Info << "Finished " << curr_job->output_tag(curr_nstruct) << " in " << (pdb_end_time - pdb_start_time) << " seconds." << std::endl;
		abinitio_protocol.get_checkpoints().clear_checkpoints();
		abrelax_checkpoints_.clear_checkpoints();
	} // end of production loop

	jobdist.estimation_of_distribution(prot_ptr);
	jobdist.reload( input_jobs );
		
	}//End of iteration X
	//	prot_ptr->print_probas();
	jobdist.shutdown();
	
	
}


void EdaAbrelax::run() {
	using namespace basic::options::OptionKeys;
	setup();
	if ( option [ OptionKeys::abinitio::rerun ] ) {
		do_rerun();
		return;
	}

	if ( option [ OptionKeys::abinitio::jdist_rerun ] ) {
		do_distributed_rerun();
		return;
	}

	// setup pose and abinitio
	EdaAbinitioOP prot_ptr;
	EdaPose init_pose;
#ifdef BOINC
	std::cerr << "Setting up folding (abrelax) ..." << std::endl;
#endif
	setup_fold( init_pose, prot_ptr ); //init_pose may or may not be fullatom (depending on flag option[ in:file:fullatom] )
#ifdef BOINC
	std::cerr << "Beginning folding (abrelax) ... " << std::endl;
#endif
	fold( init_pose, prot_ptr );
	return;
}


}
}
