// -*- mode:c++;tab-width:2;indent-tabs-mode:t;show-trailing-whitespace:t;rm-trailing-spaces:t -*-
// vi: set ts=2 noet:
// :noTabs=false:tabSize=4:indentSize=4:
//
// (c) Copyright Rosetta Commons Member Institutions.
// (c) This file is part of the Rosetta software suite and is made available under license.
// (c) The Rosetta software is developed by the contributing members of the Rosetta Commons.
// (c) For more information, see http://www.rosettacommons.org. Questions about this can be
// (c) addressed to University of Washington UW TechTransfer, email: license@u.washington.edu.

/// @file   EdaJobDistributor.cc
/// @brief
/// @author David Simoncini

// #include <utility/pointer/boost/owning_ptr.hh>
// #include <utility/pointer/boost/access_ptr.hh>
// #include <utility/pointer/boost/ReferenceCount.hh>

#include <utility/io/izstream.hh>
#include <protocols/jobdist/JobDistributors.hh>
#include "EdaJobDistributor.hh"
#include "PdbDesc.hh"
#include <basic/Tracer.hh>

#include <core/svn_version.hh>
#include <core/types.hh>
//#include <core/io/pdb/file_data.hh>
#include <core/import_pose/atom_tree_diffs/atom_tree_diff.hh>
#include <core/conformation/Residue.hh>
#include <basic/options/option.hh>
#include <core/pose/Pose.hh>
#include "EdaPose.hh"
#include "EdaAbinitio.hh"
#include <core/scoring/Energies.hh>
#include <core/scoring/ScoreFunction.hh>
// AUTO-REMOVED #include <core/scoring/ScoreFunction.hh>
// AUTO-REMOVED #include <core/scoring/ScoreFunctionFactory.hh>

#include <core/io/silent/util.hh>
#include <core/io/raw_data/DecoyFileData.hh>
#include <core/io/raw_data/ScoreFileData.hh>
#include <core/io/silent/SilentFileData.hh>
#include <core/io/silent/SilentFileData.fwd.hh>
#include <core/io/silent/SilentStructFactory.hh>

#include <numeric/random/random.hh>
#include <numeric/numeric.functions.hh>

#include <utility/exit.hh>
#include <utility/file/file_sys_util.hh>
#include <utility/file/FileName.hh>
#include <utility/io/ozstream.hh>
#include <utility/vector1.hh>
#include <utility/pointer/ReferenceCount.hh>
#ifdef USEMPI
#include <utility/string_util.hh>
#endif

// ObjexxFCL headers
#include <ObjexxFCL/string.functions.hh>

#include <map>
#include <set>
#include <sstream>
#include <string>


// option key includes

#include <basic/options/keys/out.OptionKeys.gen.hh>
#include <basic/options/keys/run.OptionKeys.gen.hh>
#include <basic/options/keys/eda.OptionKeys.gen.hh>

#include <protocols/checkpoint/Checkpoint.hh>
#include <numeric/random/random.fwd.hh>
#include "durandal/Singleton.h"
#include "durandal/DistMatrix.h"
#include "durandal/SimpPDB.h"

using namespace std;
using namespace basic::options;
using namespace basic::options::OptionKeys;

static basic::Tracer tr("EdaJobDistributor");

void print_cluster(ostream& out,
                   vector<int>& cluster,
                   vector<int>& remaining_pdbs,
                   DistMatrix& dm,
                   bool verbose,
                   bool pole_position,
                   bool compute_D) {

  int elt = cluster[0];
  bool has_energy = dm.has_energies();

  if (has_energy) {
    dm.display_energy(out, cluster);
  }
  if (compute_D) {
    out << "D:    " << dm.compute_D(cluster) << endl;
  }
  out << "members(" << cluster.size() << "):";
  if (verbose) {
    out << '\n' << dm.resolve(elt);
    if (has_energy) {
      for (size_t energy_index = 0 ; energy_index < dm.get_nb_energies();
           ++energy_index) {
        out << ':' << dm.resolve_energy(elt, energy_index);
      }
    }
  } else {
    out << ' ' << elt;
  }
  if (not pole_position) {
    remove(remaining_pdbs, elt);
  }
  for (size_t i = 1 ; i < cluster.size() ; ++i) {
    elt = cluster[i];
    if (verbose) {
      out << '\n' << dm.resolve(elt);
      if (has_energy) {
        for (size_t energy_index = 0 ; energy_index < dm.get_nb_energies();
             ++energy_index) {
          out << ':' << dm.resolve_energy(elt, energy_index);
        }
      }
    } else {
      out << ' ' << elt;
    }
    if (not pole_position) {
      remove(remaining_pdbs, elt);
    }
  }
  out << '\n';
}

void output_cluster(ostream& out,
                    vector<int>& cluster,
                    vector<int>& remaining_pdbs,
                    vector< vector<int> >& pole_position_clusters,
                    DistMatrix& dm,
                    bool verbose ,
                    bool stable,
                    bool compute_D) {

  if (pole_position_clusters.size() > 0) {
    if (stable) {
      out << "<pole position clusters("
          << pole_position_clusters.size() << ")>:\n";
      for (size_t i = 0 ; i < pole_position_clusters.size() ; ++i) {
        print_cluster(out, pole_position_clusters[i], remaining_pdbs,
                      dm, verbose, true, compute_D);
      }
      out << "</pole position clusters>\n";
    } else {
      out << "pole position centers("
          << pole_position_clusters.size() << "):";
      if (verbose) {
        out << '\n';
        for (size_t i = 0 ; i < pole_position_clusters.size() ; ++i) {
          out << dm.resolve(pole_position_clusters[i][0]) << '\n';
        }
      } else {
        for (size_t i = 0 ; i < pole_position_clusters.size() ; ++i) {
          out << ' ' << pole_position_clusters[i][0];
        }
        out << '\n';
      }
    }
  }
  print_cluster(out, cluster, remaining_pdbs, dm, verbose, false, compute_D);
}


EdaJobDistributor::EdaJobDistributor(JobVector jobs):
	Parent(jobs)
{
	  Single& single = Single::instance();
		single._compute_D        = false;
		single._only_rank        = false;
		single._smart_mode       = true;
		single._verbose          = true;
		single._brute_mode       = false;
		single._output_entropy   = false;
		single._auto_switch      = true;
		single._smart_fix        = false;
		single._brute_fix        = false;
		single._no_lower_bound   = false;
		single._no_upper_bound   = false;
		single._output_to_file   = false;
		single._use_URMSD        = false;
		single._snapshot_entropy = false;
		single._use_theobald     = true;
		single._reference_choosing_strategy = RANDOM;
		median_score = 100000;
		inverse_proba = false;
}


EdaJobDistributor::~EdaJobDistributor() {}

void EdaJobDistributor::store_score_and_frags(EdaPose & pose, string outtag, core::scoring::ScoreFunction const & scorefxn) {
	scores_.push_back(scorefxn(pose));
	fragkeys_list_.push_back(pose.get_fragkeys());
	PdbDesc _pdb;
	_pdb.name = outtag;
	for (int i=1;i<=pose.total_residue();i++)
		{
			core::conformation::Residue rsd = pose.residue(i);
			for ( Size j = 1; j <= rsd.natoms(); ++j ) 
				{
					conformation::Atom atom( rsd.atom(j) );
					if (rsd.atom_name(j)==" CA ")
						{
							_pdb.coords.push_back((float) atom.xyz()(1));
							_pdb.coords.push_back((float) atom.xyz()(2));
							_pdb.coords.push_back((float) atom.xyz()(3));
							continue;
						}
				}
		}
	if (!_pdb.coords.size())
		{
			cerr << "Error in EdaJobDistributor: No coordinates in pdb sent to master" << endl; 
			exit(1);
		}
	pdbdescs_.push_back(_pdb);
}


void EdaJobDistributor::reload(JobVector jobs) {
	jobs_=jobs;
	current_job_=1;
	scores_.clear();
	if (mpi_rank()!= 0) 
		fragkeys_list_.clear();
	pdbdescs_.clear();
}

void EdaJobDistributor::startup() {

 	// calling base class startup
 	Parent::startup();
	if (option[ OptionKeys::eda::nb_iterations ]()*mpi_nprocs() > option[ out::nstruct ])
		{
			cout << "Error: nstruct must be at least as big as nb_procs * nb_iterations" << endl;
			exit(1);
		}
}


#ifdef USEMPI

int EdaJobDistributor::get_mpi_rank() {
	return mpi_rank();
}

void EdaJobDistributor::ship_scores() {
	MPI_Datatype score_vec;
	int vec_size = scores_.size();
	MPI_Type_contiguous(vec_size, MPI_DOUBLE, &score_vec);
	MPI_Type_commit(&score_vec);
	MPI_Send(&vec_size, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
	MPI_Send(&scores_.front(), 1, score_vec, 0, 0, MPI_COMM_WORLD);
	MPI_Type_free(&score_vec);
}

utility::vector1< utility::vector1< double > > EdaJobDistributor::receive_scored_pdbs() 
{
	utility::vector1< vector< vector <float> > > all_coords;
	utility::vector1< vector< vector <char> > > all_names;
	utility::vector1< utility::vector1< double > > all_scores;
	int how_many=mpi_nprocs()-1;
	all_coords.resize(how_many);
	all_names.resize(how_many);
	all_scores.resize(how_many);
	for (unsigned i=1;i<=how_many;i++) {
		MPI_Datatype score_vec;
		int vec_size;
		MPI_Recv(&vec_size, 1, MPI_INT, MPI_ANY_SOURCE, 0, MPI_COMM_WORLD, &stat_);
		MPI_Type_contiguous(vec_size, MPI_DOUBLE, &score_vec);
		MPI_Type_commit(&score_vec);
		all_scores[stat_.MPI_SOURCE].resize(vec_size, (double) 0);
		MPI_Recv(&all_scores[stat_.MPI_SOURCE].front(), 1, score_vec, stat_.MPI_SOURCE, 0, MPI_COMM_WORLD, &stat_);
		MPI_Type_free(&score_vec);

		int nbstructs;
		MPI_Recv(&nbstructs, 1, MPI_INT, stat_.MPI_SOURCE, 0, MPI_COMM_WORLD, &stat_);
		all_coords[stat_.MPI_SOURCE].resize(nbstructs);
		all_names[stat_.MPI_SOURCE].resize(nbstructs);
		for (int j=0;j<nbstructs;j++)
			{
				MPI_Datatype coords_vec;
				MPI_Datatype char_vec;
				int vecsize;
				int stringsize;
				MPI_Recv(&stringsize, 1, MPI_INT, stat_.MPI_SOURCE, 0, MPI_COMM_WORLD, &stat_);
				MPI_Type_contiguous(stringsize, MPI_CHAR, &char_vec);
				MPI_Type_commit(&char_vec);
				all_names[stat_.MPI_SOURCE][j].resize(stringsize);
				MPI_Recv(&all_names[stat_.MPI_SOURCE][j].front(), 1, char_vec, stat_.MPI_SOURCE, 0, MPI_COMM_WORLD, &stat_);
				MPI_Recv(&vecsize, 1, MPI_INT, stat_.MPI_SOURCE, 0, MPI_COMM_WORLD, &stat_);
				MPI_Type_contiguous(vecsize, MPI_FLOAT, &coords_vec);
				MPI_Type_commit(&coords_vec);
				all_coords[stat_.MPI_SOURCE][j].resize(vecsize, (float) 0);
				MPI_Recv(&all_coords[stat_.MPI_SOURCE][j].front(), 1, coords_vec, stat_.MPI_SOURCE, 0, MPI_COMM_WORLD, &stat_);
				MPI_Type_free(&coords_vec);
				MPI_Type_free(&char_vec);
			}
	}
	for (unsigned i=1;i<=how_many;i++) {
		for (unsigned j=1;j<=all_scores[i].size();j++)
			scores_.push_back(all_scores[i][j]);
	}

	for (unsigned i=1;i<=how_many;i++) {
		for (unsigned j=0;j<all_coords[i].size();j++)
			{
				PdbDesc _pdb;
				for (unsigned k=0;k<all_coords[i][j].size();k++)
					{
						_pdb.coords.push_back(all_coords[i][j][k]);
					}
				_pdb.name.resize(all_names[i][j].size());
				for (unsigned k=0;k<all_names[i][j].size();k++)
					_pdb.name[k] = all_names[i][j][k];
				pdbdescs_.push_back(_pdb);
			}
	}
return all_scores; 
}

void EdaJobDistributor::receive_scores_and_keys(double rate) {
	if ( option[ OptionKeys::eda::clustering ]() ) {
		utility::vector1< utility::vector1< double > > all_scores;
		int how_many=mpi_nprocs()-1;

		all_scores = receive_scored_pdbs();
		vector<double> inline_scores;
		vector<IndexClust> index2job;
		for (unsigned i=1; i<=all_scores.size(); i++) {
			for (unsigned j=1;j<=all_scores[i].size();j++) {
				inline_scores.push_back(all_scores[i][j]);
				IndexClust ic;
				ic.node = i;
				ic.job_num = j;
				index2job.push_back(ic);
			}
		}
	

		vector <SimPDB*> simpdbs;
	
		for (int i=1;i<=pdbdescs_.size();i++)
			{
				SimPDB * simpdb = new SimPDB(pdbdescs_[i],false);
				simpdbs.push_back(simpdb);
			}
		clock_t start = get_user_plus_system_times();
		DistMatrix dm(simpdbs, 4, start);
		dm.smart_init();
		// Sort scores, establish threshold. 
		// 1. Collect pose indexes that are below threshold
		// 2. Send vector of pose indexes to each worker
		// 3. Receive fragment keys
		ofstream out ("clusters.dat", ios_base::app);
		if (not out.is_open()) {
			cout << "error: cannot write" << endl;
		}
		// find and print all clusters
		vector<int> remaining_pdbs = dm.get_all_PDBs_index();
		vector<IndexClust> cluster_centers;
		int lec_index;
		for (int printed = 0 ;
				 remaining_pdbs.size() > 0 and printed < 100 ;
				 ++printed) {

			vector<int> lowest_energy_cluster;
			vector< vector<int> > pole_position_clusters;

			// find biggest
			lec_index = dm.get_lowest_energy_cluster(remaining_pdbs,
																							 lowest_energy_cluster,
																							 inline_scores);
		
			cluster_centers.push_back(index2job[lec_index]);
			// output and remove from remaining
			output_cluster(out, lowest_energy_cluster, remaining_pdbs,
										 pole_position_clusters, dm, true, false,
										 false);
		}
		out.close();
		if (inline_scores.size()>=5)
			{
				sort(inline_scores.begin(), inline_scores.end());
				if (inline_scores[4]>median_score && !inverse_proba)
					inverse_proba=true;
				else
					{
						inverse_proba=false;
						fragkeys_list_.clear();
					}
				median_score = inline_scores[4];
			}
		utility::vector1 < utility::vector1 < int > > indexes;
		indexes.resize(how_many);
		for (unsigned i=0; i<cluster_centers.size(); i++) 
			{
				indexes[cluster_centers[i].node].push_back(cluster_centers[i].job_num);
			}
		for (unsigned i=1; i<=how_many; i++) {
			int vec_size=indexes[i].size();
			MPI_Datatype index_vec;
			MPI_Type_contiguous(vec_size, MPI_INT, &index_vec);
			MPI_Type_commit(&index_vec);
			MPI_Send(&vec_size, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
			MPI_Send(&indexes[i].front(), 1, index_vec, i, 0, MPI_COMM_WORLD);
			MPI_Type_free(&index_vec);
			for (unsigned j=1;j<=indexes[i].size();j++)
				receive_fragkeys();
		}
	}
	else
		{
			utility::vector1< utility::vector1< double > > all_scores;
			int how_many=mpi_nprocs()-1;
			all_scores.resize(how_many);
			for (unsigned i=1;i<=how_many;i++) {
				MPI_Datatype score_vec;
				int vec_size;
				MPI_Recv(&vec_size, 1, MPI_INT, MPI_ANY_SOURCE, 0, MPI_COMM_WORLD, &stat_);
				MPI_Type_contiguous(vec_size, MPI_DOUBLE, &score_vec);
				MPI_Type_commit(&score_vec);
				all_scores[stat_.MPI_SOURCE].resize(vec_size, (double) 0);
				MPI_Recv(&all_scores[stat_.MPI_SOURCE].front(), 1, score_vec, stat_.MPI_SOURCE, 0, MPI_COMM_WORLD, &stat_);
				MPI_Type_free(&score_vec);
				for (unsigned j=1;j<=all_scores[i].size();j++)
					scores_.push_back(all_scores[i][j]);
			}
			// Sort scores, establish threshold. 
			// 1. Collect pose indexes that are below threshold
			// 2. Send vector of pose indexes to each worker
			// 3. Receive fragment keys
			sort(scores_.begin(), scores_.end());
			double threshold_energy = scores_[(int) (scores_.size()*rate)>1?(int) (scores_.size()*rate):1]; // Just a security to prevent crash on dumb threshold value
			for (unsigned i=1; i<=how_many; i++) {
				utility::vector1< int > indexes;
				for (unsigned j=1;j<=all_scores[i].size();j++) {
					if (all_scores[i][j]<=threshold_energy)
						{
							indexes.push_back((int) j);
						}
				}
				int vec_size=indexes.size();
				MPI_Datatype index_vec;
				MPI_Type_contiguous(vec_size, MPI_INT, &index_vec);
				MPI_Type_commit(&index_vec);
				MPI_Send(&vec_size, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
				MPI_Send(&indexes.front(), 1, index_vec, i, 0, MPI_COMM_WORLD);
				MPI_Type_free(&index_vec);
				for (unsigned j=1;j<=indexes.size();j++)
					receive_fragkeys();
				indexes.clear();
			}
		}
}

void EdaJobDistributor::receive_order_and_send_keys() {
	int vec_size;
	MPI_Recv(&vec_size, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &stat_);
	utility::vector1< int > indexes;
	indexes.resize(vec_size);
	MPI_Datatype index_vec;
	MPI_Type_contiguous(vec_size, MPI_INT, &index_vec);
	MPI_Type_commit(&index_vec);
	MPI_Recv(&indexes.front(), 1, index_vec, 0, 0, MPI_COMM_WORLD, &stat_);
	for (unsigned i=1;i<=indexes.size();i++) {
		ship_fragkeys(fragkeys_list_[indexes[i]]);
	}

	MPI_Type_free(&index_vec);
}

void EdaJobDistributor::ship_pdb(PdbDesc _pdb)
{
	int vecsize = _pdb.coords.size();
	//	cout << "Inside ship_pdbs" << endl;
	MPI_Datatype coords_vec;
	MPI_Datatype char_vec;
	vector< char > str2vec;
	for(size_t i=0;i<_pdb.name.size();i++)
		{
			str2vec.push_back(_pdb.name[i]);
		}
	int charvecsize = str2vec.size();
	//	cout << "name " << _pdb.name <<  endl;
	MPI_Type_contiguous(charvecsize, MPI_CHAR, &char_vec);
	MPI_Type_commit(&char_vec);
	MPI_Send(&charvecsize, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
	MPI_Send(&str2vec.front(), 1, char_vec, 0, 0, MPI_COMM_WORLD);
	//	cout << "Bam " << vecsize << endl;
	MPI_Type_contiguous(vecsize, MPI_FLOAT, &coords_vec);
	MPI_Type_commit(&coords_vec);
	MPI_Send(&vecsize, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
	MPI_Send(&_pdb.coords.front(), 1, coords_vec, 0, 0, MPI_COMM_WORLD);
	//	cout << "Boum" << endl;
	MPI_Type_free(&char_vec);
	MPI_Type_free(&coords_vec);
}


void EdaJobDistributor::ship_fragkeys(utility::vector1< EdaPose::Item > fragKeys) {
	// Setting up stuff for communication with master.
	int nb_keys = (int) fragKeys.size();
	MPI_Datatype item, all_items;
	MPI_Datatype types[2];
	MPI_Aint disp[2];
	int block_len[2];
	block_len[0] = block_len[1] = 1;
	types[0] = MPI_INT;
	types[1] = MPI_INT;
	MPI_Address(&fragKeys.front().frame_pos, &disp[0]);
	MPI_Address(&fragKeys.front().frag_num, &disp[1]);
	disp[1] = disp[1]-disp[0];
	disp[0] = 0;
	MPI_Type_struct(2, block_len, disp, types, &item);
	MPI_Type_contiguous(nb_keys, item, &all_items);
	MPI_Type_commit(&all_items);
	MPI_Send(&nb_keys, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
	MPI_Send(&fragKeys.front(), 1, all_items, 0, 0, MPI_COMM_WORLD);
	MPI_Type_free(&all_items);
}

void EdaJobDistributor::receive_fragkeys() {
	// Setting up stuff for communication with master. SHOULD SPECIFY VEC SIZE FOR MASTER....
	utility::vector1< EdaPose::Item > fragKeys;
	int nb_keys;
	MPI_Recv(&nb_keys, 1, MPI_INT, MPI_ANY_SOURCE, 0, MPI_COMM_WORLD, &stat_);
	fragKeys.resize(nb_keys);
	MPI_Datatype item, all_items;
	MPI_Datatype types[2];
	MPI_Aint disp[2];
	int block_len[2];
	block_len[0] = block_len[1] = 1;
	types[0] = MPI_INT;
	types[1] = MPI_INT;
	MPI_Address(&fragKeys.front().frame_pos, &disp[0]);
	MPI_Address(&fragKeys.front().frag_num, &disp[1]);
	disp[1] = disp[1]-disp[0];
	disp[0] = 0;
	MPI_Type_struct(2, block_len, disp, types, &item);
	MPI_Type_contiguous(nb_keys, item, &all_items);
	MPI_Type_commit(&all_items);
	MPI_Recv(&fragKeys.front(), 1, all_items, stat_.MPI_SOURCE, 0, MPI_COMM_WORLD, &stat_);
	if (!inverse_proba)
		fragkeys_list_.push_back(fragKeys);
	MPI_Type_free(&all_items);
}


#endif


void EdaJobDistributor::estimation_of_distribution(EdaAbinitioOP& prot_ptr) {
#ifdef USEMPI
	if (get_mpi_rank()==0)
		{
			time_t shipping_start_time = time(NULL);
			receive_scores_and_keys(prot_ptr->threshold());
			// for (Size i=0;i<nstruct;i++)
			// 	jobdist.receive_fragkeys();
			time_t shipping_end_time = time(NULL);
			tr.Info << "Shipped data to master in " << shipping_end_time - shipping_start_time << " seconds." << std::endl; 
			prot_ptr->update_probas(fragkeys_list_, inverse_proba);
		}
	else {
		ship_scores();
		int nbstruct = pdbdescs_.size();
		MPI_Send(&nbstruct, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
		cout << "Sending nbstruct " << nbstruct << " from " << mpi_rank() << endl;
		for (unsigned i=1; i<=pdbdescs_.size();i++)
			ship_pdb(pdbdescs_[i]);
		receive_order_and_send_keys();
	}
	prot_ptr->notify_probas();
	return;
#endif
	
}


int EdaJobDistributor::fragkeys_list_size() {
	return fragkeys_list_.size();
}

void EdaJobDistributor::print_keys(std::ostream & output) { 
	for (unsigned i=1;i<=fragkeys_list_.size();i++)
		{
			for (unsigned j=1;j<=fragkeys_list_[i].size();j++)
				{
					output << fragkeys_list_[i][j].frame_pos << ":" << fragkeys_list_[i][j].frag_num << " ";
				}
			output << std::endl;
		}
}

utility::vector1< utility::vector1< EdaPose::Item > > & EdaJobDistributor::get_fragkeys_list() {
	return fragkeys_list_;
}

